import sys
import tempfile
import subprocess
from pathlib import Path

import click
import requests

from .nixpkgs_repo import get_repo, packages, packages_for_sha


def build_in_path(args, attrs, path):
    """Build the given package attributes in the given nixpkgs path"""
    if not attrs:
        click.echo('Nothing changed')
        return

    canonical_path = str(Path(path).resolve())
    result_dir = tempfile.mkdtemp(prefix='nox-review-')
    click.echo('Building in {}: {}'.format(click.style(result_dir, bold=True),
                                           click.style(' '.join(attrs), bold=True)))
    command = ['nix-build']
    command += args
    for a in attrs:
        command.append('-A')
        command.append(a)
    command.append(canonical_path)

    try:
        subprocess.check_call(command, cwd=result_dir)
    except subprocess.CalledProcessError:
        click.secho('The invocation of "{}" failed'.format(' '.join(command)), fg='red')
        sys.exit(1)
    click.echo('Result in {}'.format(click.style(result_dir, bold=True)))
    subprocess.check_call(['ls', '-l', result_dir])


def build_sha(args, attrs, sha):
    """Build the given package attributs for a given sha"""
    repo = get_repo()
    repo.checkout(sha)
    build_in_path(args, attrs, repo.path)


def differences(old, new):
    """Return set of attributes that changed between two packages list"""
    raw = new - old
    # Only keep the attribute name
    return {l.split()[0] for l in raw}


@click.group()
@click.option('--keep-going', '-k', is_flag=True, help='Keep going in case of failed builds')
@click.pass_context
def cli(ctx, keep_going):
    """Review a change by building the touched commits"""
    if keep_going:
        ctx.obj = {'extra-args': ['--keep-going']}
    else:
        ctx.obj = {'extra-args': []}


@cli.command(short_help='difference between working tree and a commit')
@click.option('--against', default='HEAD')
@click.pass_context
def wip(ctx, against):
    """Build in the current dir the packages that different from AGAINST (default to HEAD)"""
    if not Path('pkgs/top-level/all-packages.nix').exists():
        click.secho('"nox-review wip" must be run in a nixpkgs repository.', fg='red')
        return


    dirty_working_tree = subprocess.call('git diff --quiet --ignore-submodules HEAD'.split())

    if not dirty_working_tree:
        if against == 'HEAD':
            click.secho('No uncommit changes. Did you mean to use the "--against" option?')
            return


    sha = subprocess.check_output(['git', 'rev-parse', '--verify', against]).decode().strip()

    attrs = differences(packages_for_sha(sha),
                        packages('.'))

    build_in_path(ctx.obj['extra-args'], attrs, '.')


@cli.command('pr', short_help='changes in a pull request')
@click.argument('pr', type=click.INT)
@click.pass_context
def review_pr(ctx, pr):
    """Build the changes induced by the given pull request"""
    payload = requests.get('https://api.github.com/repos/NixOS/nixpkgs/pulls/{}'.format(pr)).json()
    click.echo('=== Reviewing PR {} : {}'.format(click.style(str(pr), bold=True),
                                             click.style(payload['title'], bold=True)))

    base_ref = payload['base']['ref']

    repo = get_repo()

    click.echo('==> Fetching base ({})'.format(base_ref))
    base_refspec = 'heads/{}'.format(payload['base']['ref'])
    repo.fetch(base_refspec)
    base = repo.sha('FETCH_HEAD')

    click.echo('==> Fetching PR')
    head_refspec = 'pull/{}/head'.format(pr)
    repo.fetch(head_refspec)
    head = repo.sha('FETCH_HEAD')

    click.echo('==> Fetching extra history for merging')
    depth = 10
    while not repo.merge_base(head, base):
        repo.fetch(base_refspec, depth=depth)
        repo.fetch(head_refspec, depth=depth)
        depth *=2

    # It looks like this isn't enough for a merge, so we fetch more
    repo.fetch(base_refspec, depth=depth)

    click.echo('==> Merging PR into base')
    repo.checkout(base)
    repo.git(['merge', head, '-qm', 'Nox automatic merge'])
    merged = repo.sha('HEAD')

    attrs = differences(packages_for_sha(base),
                        packages_for_sha(merged))

    build_sha(ctx.obj['extra-args'], attrs, merged)



import click
import collections
import re
import subprocess

from pathlib import Path
from characteristic import attributes
from .cache import region


@region.cache_on_arguments()
def query(*args):
    return subprocess.check_output(['nix-store', '--query'] + list(args),
                                   universal_newlines=True)


def requisites(drv):
    return set(query('--requisites', drv).strip().split('\n'))


def is_blacklisted(path):
    # TODO look in the derivation to known if it's a fetchurl
    # TODO : check what those 'stage-*' are
    return any(s in path for s in ('.tar.', '.tgz', '.zip', 'stage-1-init', 'stage-2-init', '.patch'))


@attributes(['path'], apply_with_init=False)
class NixPath:
    def __init__(self, path):
        self.path = path
        self.is_drv = self.path.endswith('.drv')
        name_slice_end = -4 if self.is_drv else None
        self.full_name = self.path[44:name_slice_end]

        m = re.search(r'-(\d.*)', self.full_name)
        if m:
            self.name = self.full_name[:m.start()]
            self.version = self.full_name[m.start()+1:]
        else:
            self.name = self.full_name
            self.version = None

    def reqs(self):
        result = requisites(self.path)
        result.discard(self.path)
        return {NixPath(p) for p in result if not is_blacklisted(p)}

    def refs(self):
        return set(query('--references', self.path).strip().split('\n'))


def current_system_drv():
    current_system = Path('/run/current-system').resolve()
    return NixPath(query('--deriver', str(current_system)).strip())


def new_system_drv():
    with subprocess.Popen(['nixos-rebuild', 'dry-run'],
                          stderr=subprocess.PIPE,
                          universal_newlines=True) as process:
        _, output = process.communicate()

    m = re.search(r'.*nixos-\d{2}.*', output)
    return m and NixPath(m.group().strip())


def display_path(path, bold):
    is_drv = path.endswith('.drv')
    name_slice_end = -4 if is_drv else None
    return (path[:44] +
            click.style(path[44:name_slice_end], bold=bold) +
            (path[name_slice_end:] if is_drv else ''))


class DepsTree:
    def __init__(self, changed_refs_tree):
        self.seen = set()
        self.changed_refs_tree = changed_refs_tree

    def show(self, pkg, level=0):
        click.echo('  '*level + display_path(pkg, bold=pkg not in self.seen), nl=False)
        if pkg not in self.seen:
            click.echo()
            self.seen.add(pkg)
            for ref in self.changed_refs_tree[pkg]:
                self.show(ref, level+1)
        else:
            click.echo(' [...]')


@click.command()
def main():
    new_drv = new_system_drv()
    if not new_drv:
        click.echo('No system updates')
        return

    current_drv = current_system_drv()

    # must differenciate firefox-31 and firefox
    current_names = {(drv.name, bool(drv.version)): drv for drv in current_drv.reqs()
                     if not is_blacklisted(drv.path)}
    changed_refs_tree = collections.defaultdict(list)

    new_packages, new_versions, changed_expressions, rest = [], [], [], []

    for pkg in new_drv.reqs() - current_drv.reqs():
        previous = current_names.get((pkg.name, bool(pkg.version)))
        if not previous:
            new_packages.append(pkg)
            continue

        changed_refs = pkg.refs() - previous.refs()
        for r in changed_refs:
            changed_refs_tree[r].append(pkg.path)

        if previous.version != pkg.version:
            new_versions.append((pkg, previous.version))
        elif previous.refs() == pkg.refs():
            changed_expressions.append(pkg)
        else:
            rest.append(pkg)

    new_versions.sort(key=lambda x: x[0].name)
    for l in (new_packages, changed_expressions, rest):
        l.sort(key=lambda p: p.name)

    tree = DepsTree(changed_refs_tree)

    for pkg in new_packages:
        click.secho('{} : seems to be new'.format(pkg.full_name), bold=True)
        tree.show(pkg.path)
    for pkg, old_version in new_versions:
        click.secho('{} : new version ({} -> {})'.format(pkg.name,
                                                         old_version,
                                                         pkg.version),
                    bold=True)
        tree.show(pkg.path)
    for pkg in changed_expressions:
        click.secho('{} : new expression'.format(pkg.full_name), bold=True)
        tree.show(pkg.path)

    forgotten = set(p.path for p in rest) - tree.seen
    if forgotten:
        print('Some packages where forgotten : {}'.format(forgotten))

# TODO : option -> display only thing to install, deps tree without
# repetition, deps tree with omission of repeated paths, or full tree



import os
import collections
import json
import subprocess

import click

from .cache import region


class NixEvalError(Exception):
    pass


def nix_packages_json():
    click.echo('Refreshing cache')
    try:
        output = subprocess.check_output(['nix-env', '-qa', '--json', '--show-trace'],
                                         universal_newlines=True)
    except subprocess.CalledProcessError as e:
        raise NixEvalError from e
    return json.loads(output)


Package = collections.namedtuple('Package', 'attribute name description')


def key_for_path(path):
    try:
        manifest = os.path.join(path, 'manifest.nix')
        with open(manifest) as f:
            return f.read()
    except (FileNotFoundError, NotADirectoryError):
        pass
    if os.path.exists(os.path.join(path, '.git')):
        return subprocess.check_output('git rev-parse --verify HEAD'.split(),
                                       cwd=path)
    click.echo('Warning: could not find a version indicator for {}'.format(path))
    return None


def all_packages():
    defexpr = os.path.expanduser('~/.nix-defexpr/')
    paths = os.listdir(defexpr)
    key = str({p: key_for_path(defexpr + p) for p in paths})
    packages_json = region.get_or_create(key, nix_packages_json)
    return (Package(attr, v['name'], v['meta'].get('description', ''))
            for attr, v in packages_json.items())


@click.command()
@click.argument('query', default='')
def main(query):
    """Search a package in nix"""
    try:
        results = [p for p in all_packages()
                   if any(query in s for s in p)]
    except NixEvalError:
        raise click.ClickException('An error occured while running nix (displayed above). Maybe the nixpkgs eval is broken.')
    results.sort()
    for i, p in enumerate(results, 1):
        line = '{} {} ({})\n    {}'.format(
            click.style(str(i), fg='black', bg='yellow'),
            click.style(p.name, bold=True),
            click.style(p.attribute, dim=True),
            click.style(p.description))
        click.echo(line)

    if results:
        def parse_input(inp):
            if inp[0] == 's':
                action = 'shell'
                inp = inp[1:]
            else:
                action = 'install'
            packages = [results[int(i) - 1] for i in inp.split()]
            return action, packages

        action, packages = click.prompt('Packages to install',
                                        value_proc=parse_input)
        attributes = [p.attribute for p in packages]
        if action == 'install':
            subprocess.check_call(['nix-env', '-iA', '--show-trace'] + attributes)
        elif action == 'shell':
            attributes = [a[len('nixpkgs.'):] for a in attributes]
            subprocess.check_call(['nix-shell', '-p', '--show-trace'] + attributes)



import os
import subprocess
from pathlib import Path

from .cache import region

import click


class Repo:
    def __init__(self):
        nox_dir = Path(click.get_app_dir('nox', force_posix=True))
        if not nox_dir.exists():
            nox_dir.mkdir()

        nixpkgs = nox_dir / 'nixpkgs'
        self.path = str(nixpkgs)

        if not nixpkgs.exists():
            click.echo('==> Creating nixpkgs repo in {}'.format(nixpkgs))
            self.git(['init', '--quiet', self.path], cwd=False)
            self.git('remote add origin https://github.com/NixOS/nixpkgs.git')
            self.git('config user.email nox@example.com')
            self.git('config user.name nox')


        if (Path.cwd() / '.git').exists():
            git_version = self.git('version', output=True).strip()
            if git_version >= 'git version 2':
                click.echo("==> We're in a git repo, trying to fetch it")

                self.git(['fetch', str(Path.cwd()), '--update-shallow', '--quiet'])
            else:
                click.echo("==> Old version of git detected ({}, maybe on travis),"
                " not trying to fetch from local, fetch 50 commits from master"
                " instead".format(git_version))
                self.git('fetch origin master --depth 50')

    def git(self, command, *args, cwd=None, output=False, **kwargs):
        if cwd is None:
            cwd = self.path
        elif cwd is False:
            cwd = None
        if isinstance(command, str):
            command = command.split()
        command.insert(0, 'git')
        f = subprocess.check_output if output else subprocess.check_call
        return f(command, *args, cwd=cwd, universal_newlines=output, **kwargs)




    def checkout(self, sha):
        self.git(['checkout', '--quiet', sha])

    def sha(self, ref):
        return self.git(['rev-parse', '--verify', ref], output=True).strip()

    def fetch(self, ref, depth=1):
        return self.git(['fetch', '--depth', str(depth), '--quiet',
            'origin', '+refs/{}'.format(ref)])

    def merge_base(self, first, second):
        try:
            return self.git(['merge-base', first, second], output=True).strip()
        except subprocess.CalledProcessError:
            return None

_repo = None

def get_repo():
    global _repo
    if not _repo:
        _repo = Repo()
    return _repo


def packages(path):
    """List all nix packages in the repo, as a set"""
    output = subprocess.check_output(['nix-env', '-f', path, '-qaP', '--drv-path', '--show-trace'],
                                     universal_newlines=True)
    return set(output.split('\n'))


@region.cache_on_arguments()
def packages_for_sha(sha):
    """List all nix packages for the given sha"""
    repo = get_repo()
    repo.checkout(sha)
    return packages(repo.path)
